
/*
Dirk's Bot abilites
	by Lach and CobaltBW

Ported and moved from 2.1 Inazuma
Assistance from SwitchKaze
*/

local debug = false -- Set to true to output the botticcmd timeline
local dprint = function(str)
	return debug and print(str)
end

-- Dirk AI (ThinkState controller)
local setThink = function(ai, str, angle)
	str = $ or 'default'
	assert(type(str) == 'string', 'setThink argument #2 must be a string! Got argument "'..tostring(str)..'"')
	dprint('setting thinkstate from '..tostring(ai.thinkstate..' to '..str))
	ai.thinkstate = str
	ai.thinktime = 0
	ai.catchup_tics = 0
	if angle != nil
		ai.thinkangle = angle
	end
end

-- ThinkHelp radius
local thinkHelpRadius = function(bmo, pmo, radiusonly)
	return (radiusonly or (bmo.player.speed == 0 and P_IsObjectOnGround(bmo) and bmo.player.pflags & (PF_STARTDASH|PF_SPINNING) == 0) )
		and P_AproxDistance(bmo.z - pmo.z, P_AproxDistance(bmo.x - pmo.x, bmo.y - pmo.y)) <= bmo.radius*2
end


-- Look at player
local lookAtPlayer = function(bot, player)
	bot.cmd.angleturn = R_PointToAngle2(bot.mo.x, bot.mo.y, player.mo.x, player.mo.y) >> 16
end

-- Dirk AI (thinkstate functions)
local aiFunc = {
	 -- Standard follow code. This code also gets the AI to the alternative thinkstates.
	default = function(bot, cmd, ai, player)
		local mo = player.mo
		if P_IsObjectOnGround(bot.mo)
		and thinkHelpRadius(bot.mo, mo)
			-- Set thinkhelp state
			setThink(ai, 'thinkhelp')
			return true
		end
		if (ai.catchup_tics >= 70)
			setThink(ai, 'catchup')
			return true
		end
		-- Fallthrough to vanilla AI
		return false
	end,

	-- Thinking of helping bot leader
	thinkhelp = function(bot, cmd, ai, player)
		local mo = player.mo
		
		-- Not in radius
		if not thinkHelpRadius(bot.mo, mo)
			setThink(ai) -- Reset to default thinkstate
			return false
		end
		-- Determine if we're in a position where Dirk should use his jump ability
		if player.cmd.buttons & BT_JUMP
		and not (player.pflags & PF_THOKKED)
		and player.pflags & PF_JUMPED
		and P_MobjFlip(mo) * mo.momz > 0
		and P_IsObjectOnGround(bot.mo)
		and bot.mo.iceplatforms
			setThink(ai, 'ability', mo.angle) -- Setting the angle to force the ai to look forwards
			-- Set inputs
			cmd.buttons = $ | BT_JUMP
			cmd.angleturn = ai.thinkangle >> 16
			-- This is a bit of a cheat, but it gets the job done I guess
			bot.mo.angle = cmd.angleturn << 16
			P_InstaThrust(bot.mo, bot.mo.angle, 6*bot.mo.scale)
			return true
		end
		-- Determine if we're in a position where Dirk should use his spin ability
		if player.cmd.buttons & BT_SPIN
			setThink(ai, 'ability2')
			return true
		end
		lookAtPlayer(bot, player)
		return true
	end,
	

	-- Jump ability code	
	ability = function(bot, cmd, ai, player)
		local mo = player.mo
		if bot.mo.standonice -- We've created an ice platform
			setThink(ai, 'ability_wait') -- Make us wait on this platform.
			return true
		end
		if not (bot.mo.standonice and bot.mo.iceplatforms)
			dprint('jump')
			if abs(bot.mo.momz) > 4*mo.scale/(mo.eflags & MFE_UNDERWATER and 3 or 1) -- If we're going fast enough up or down, press jump. The small gap inbetween means we will be releasing jump before pressing it again.
				cmd.buttons = $ | BT_JUMP
			else
				cmd.buttons = $ & ~BT_JUMP
			end
			return true
		end
	end,
	
	-- Jump ability part 2: Waiting on an ice platform
	ability_wait = function(bot, cmd, ai, player)
		local mo = player.mo
		if thinkHelpRadius(bot.mo, player.mo) -- Player is close enough to be helped again
		or not(bot.mo.standonice) -- Bot fell off their platform
			setThink(ai) -- Reset thinkstate
		end
		-- Fallthrough code
		lookAtPlayer(bot, player)
		return true -- Stand still
	end,
	
	-- Spin ability code
	ability2 = function(bot, cmd, ai, player)
		local mo = player.mo
		
-- 		-- Set direction
		if thinkHelpRadius(bot.mo, player.mo, true) -- Close to player
			cmd.angleturn = player.cmd.angleturn -- Player controls spindash direction
		else -- Far from player
			lookAtPlayer(bot, player) -- Bot spindashes into player
		end
		bot.mo.angle = cmd.angleturn << 16 -- Necessary hack, apparently
		bot.drawangle = bot.mo.angle -- Also a hack
		
		-- Check status
		if ai.thinktime > 1 and not(bot.pflags & (PF_SPINNING | PF_STARTDASH)) -- Something must have gone wrong if we're not in a spin state after the first think tic
			dprint('not spinning')
			setThink(ai) -- Reset to default
			return true
		end
		local releasetime
		-- Charge spindash
		if ai.thinktime > 0 and ai.thinktime < 10
			dprint('Charging spindash')
			cmd.buttons = $ | BT_SPIN
			return true
		end
		-- Release spindash
		dprint('Released spindash')
		cmd.buttons = $&~BT_SPIN
		cmd.buttons = $&~BT_JUMP
		setThink(ai, 'ability2_surf', mo.angle)
		return true
	end,
	
	-- Spin ability part 2: Surfing
	ability2_surf = function(bot, cmd, ai, player)
		local surfing = bot.mo.state == S_PLAY_SURF -- Not the most proper check, but other methods don't seem to work here.
		local spinning = bot.pflags & (PF_SPINNING|PF_STARTDASH)
		
		if not (surfing or spinning) and ai.thinktime > 10
			dprint('not a surf or spin state')
			setThink(ai) -- Reset thinkstate
			return true
		end
		-- Pre-surf state
		if not surfing
		and spinning
			-- Do spin tapping
			if ai.thinktime & 2
				dprint('press spin')
				cmd.buttons = $ | BT_SPIN
			else
				dprint('release spin')
				cmd.buttons = $ &~ BT_SPIN
			end
			lookAtPlayer(bot,player)
			cmd.forwardmove = 50 -- Needed to activate surf
			cmd.angleturn = ai.thinkangle >> 16
			return true
		end
		-- Surf state
		dprint('surfing')
		cmd.buttons = $ | BT_SPIN
		
		
		if not player.thomasandhisfriends
			-- Seek the player
			lookAtPlayer(bot, player)
			cmd.forwardmove = 50
		elseif player.precmd
			local pc = player.precmd
			-- Leader controls the surf!
			cmd.angleturn = pc.angleturn - pc.sidemove * (ANG1 >> 16) -- Copies player angleturn and applies strafe input to affect surf direction.
			if pc.forwardmove > 25
				-- Rise if the player holds forward
				cmd.buttons = $|BT_JUMP
			end
			-- By default, move at max speed to keep the surf going
			cmd.forwardmove = 50
			if pc.forwardmove < 0
				-- Let the backwards key serve as the "brakes"
				cmd.forwardmove = $ + (pc.forwardmove*2)
			end
		end
		return true
	end,
	
	-- Catching up with the ice surf
	catchup = function(bot, cmd, ai, player)
		local mo = player.mo

		local surfing = bot.mo.state == S_PLAY_SURF -- Not the most proper check, but other methods don't seem to work here.
		local spinning = bot.pflags & (PF_SPINNING|PF_STARTDASH) 
		
		if not (surfing or spinning) and ai.thinktime > 10
			dprint('not a surf or spin state')
			setThink(ai) -- Reset thinkstate
			return true
		end
		
		if ai.thinktime > 0 and not spinning
			cmd.buttons = $ | BT_SPIN
			return true
		end
		
		lookAtPlayer(bot,player)
		cmd.forwardmove = 50 -- Needed to activate surf
		
		if spinning
			-- Do spin tapping
			if ai.thinktime & 2
				dprint('press spin')
				cmd.buttons = $ | BT_SPIN
				cmd.buttons = $ &~ BT_JUMP
			else
				dprint('release spin')
				cmd.buttons = $ &~ BT_SPIN
				cmd.buttons = $ &~ BT_JUMP
				setThink(ai, 'catchup_surf')
			end
			return true
		end
		return true
	end,
	
	-- Now we're actually ice surfing
	catchup_surf = function(bot, cmd, ai, player)
		local mo = player.mo
	
		local surfing = bot.mo.state == S_PLAY_SURF -- Not the most proper check, but other methods don't seem to work here.
		local spinning = bot.pflags & (PF_SPINNING|PF_STARTDASH) 
		
		local  flip = P_MobjFlip(bot.mo)
		local  zdist = flip * (mo.z - bot.mo.z)
		
		if not (surfing or spinning) and ai.thinktime > 10
			dprint('not a surf or spin state')
			setThink(ai) -- Reset thinkstate
			return true
		end
		
		-- Pre-surf state
		if not surfing
		and spinning
			-- Do spin tapping
			if ai.thinktime & 2
				dprint('press spin')
				cmd.buttons = $ | BT_SPIN
				cmd.buttons = $ &~ BT_JUMP
			else
				dprint('release spin')
				cmd.buttons = $ &~ BT_SPIN
				cmd.buttons = $ &~ BT_JUMP
			end
			lookAtPlayer(bot,player)
			cmd.forwardmove = 50 -- Needed to activate surf
			return true
		end
		-- Surf state
		dprint('surfing')
		cmd.buttons = $ | BT_SPIN
		if surfing
			lookAtPlayer(bot,player)
			cmd.forwardmove = 50
			if zdist > 32*bot.mo.scale
				if bot.mo.z < mo.z
					cmd.buttons = $ | BT_JUMP
				else
					cmd.buttons = $ & ~BT_JUMP
				end
				return true
			end
		end
		
		if player.thomasandhisfriends
			setThink(ai, 'ability2_surf', mo.angle)
		end
		return true
	end
}

-- This hook sets player.precmd (necessary for surf controls)
addHook('PreThinkFrame',do
	for player in players.iterate do
		player.precmd = $ or {}
		local cmd = player.cmd
		local pc = player.precmd
		pc.buttons = cmd.buttons
		pc.forwardmove = cmd.forwardmove
		pc.sidemove = cmd.sidemove
		pc.angleturn = cmd.angleturn
		pc.aiming = cmd.aiming
	end
end)

local warning = false
-- Dirk AI (Master level)
addHook("BotTiccmd", function(bot,cmd)
	if paused or gamestate != GS_LEVEL
	or bot.mo == nil -- Must be playing
	or skins[bot.skin].name != 'dirk' -- Dirk only
	or not (bot.botleader and bot.botleader.mo) -- Must have a botleader
	or bot.botleader.exiting -- Must not be the end of the level
	or P_PlayerInPain(bot) or bot.playerstate != PST_LIVE -- Hurt and dead bots can't think
		bot.dirkai = nil -- Unset Dirk AI vars
		return
	end	

	-- Initialize Dirk AI vars
	bot.dirkai = $ or {
		thinkstate = 'default',
		thinktime = -1,
		catchup_tics = 0,
		thinkangle = angle,
		errorstate = false,
	}
	local ai = bot.dirkai
	
	local mo = bot.botleader.mo
	local flip = P_MobjFlip(bot.mo)
	local dist = P_AproxDistance(mo.x - bot.mo.x, mo.y - bot.mo.y)
	local zdist = flip * (mo.z - bot.mo.z)
	local bmom = P_AproxDistance(bot.mo.momx, bot.mo.momy)
	local followmax = 128 * 8 * bot.mo.scale -- Max follow distance before AI begins to enter catchup state
	local followthres = 92 * bot.mo.scale -- Distance that AI will try to reach
	local comfortheight = 96 * bot.mo.scale
	local stalled = (bmom < bot.mo.scale >> 1) and dist > followthres -- Helps to see if the AI is having trouble catching up
	
	if ai.errorstate and warning == false -- The script should never be broken past this point. If it is, this will print a message to warn the user
		warning = true
		print('\x82\ Please check console/logs for errors associated with Dirk AI.')
		debug = false -- Disable debug print messages past this point.
	end
	
	-- Update Vars
	ai.thinkstate = $ or 'default' -- This can never be nil
	ai.thinktime = $+1
	ai.errorstate = true -- The end of the function needs to be reached to reset this to false.
	
	if (dist > followmax or zdist > comfortheight or stalled)	
		ai.catchup_tics = min($+2, 70)
	else
		ai.catchup_tics = max($-1, 0)
	end

	dprint('\x81\ dirk bot\x80 thinkstate '..tostring(ai.thinkstate)..' (time: '..tostring(ai.thinktime)..')')
		
	-- Perform thinkstate's function
	local thinkFunc = aiFunc[ai.thinkstate]
	assert(type(thinkFunc) == "function", "No think function for Dirk AI thinkstate '"..tostring(ai.thinkstate).."' (got "..tostring(thinkFunc)..")")
	local returnVal = thinkFunc(bot, cmd, ai, bot.botleader) -- The return value will control whether we will fallthrough to SRB2's native TailsAI code.
	
	local player = bot.botleader
	local pmo = player.mo		


	-- Helper indicator
	if bot.mo.dirkind
		-- Display indicator
		if ai.thinkstate == 'thinkhelp'
		and bot.mo.iceplatforms
			dprint('dirk indicator (draw)')
			bot.mo.dirkind.flags2 = $ & ~MF2_DONTDRAW
		else -- Do not display indicator
			dprint("dirk indicator (don't draw)")
			bot.mo.dirkind.flags2 = $|MF2_DONTDRAW
		end
	else
		dprint('initialize dirk indicator')
		local ind = P_SpawnMobj(bot.mo.x, bot.mo.y, bot.mo.z, MT_OVERLAY)
		ind.target = bot.mo
		ind.state = S_FLIGHTINDICATOR
		ind.flags2 = $|MF2_DONTDRAW
		bot.mo.dirkind = ind
	end
	


	/*if mo.target and mo.target.valid and mo.target.flags & MF_BOSS
		bot.boss = mo.target
		bot.weapondelay = 4*TICRATE
	end

	if bot.boss and bot.boss.valid and bot.boss.health and not (bot.boss.flags2 & MF2_FRET) and not bot.weapondelay and not P_PlayerInPain(bot)
		//local ringtype = P_RandomChance(FRACUNIT/4) and MT_THROWNICE or P_RandomRange(325,329)
		bot.mo.angle = R_PointToAngle2(bot.mo.x, bot.mo.y, bot.boss.x, bot.boss.y)
		bot.aiming = R_PointToAngle2(0,bot.mo.z,P_AproxDistance(bot.mo.x-bot.boss.x,bot.mo.y-bot.boss.y),bot.boss.z)
		local ring = P_SpawnMobj(bot.mo.x,bot.mo.y,bot.mo.z+bot.mo.height/2, MT_THROWNICE)
		ring.target = bot.mo
		ring.momz = FixedMul(FixedMul(ring.info.speed,bot.mo.scale),sin(bot.aiming))
		P_InstaThrust(ring,bot.mo.angle,FixedMul(FixedMul(ring.info.speed,bot.mo.scale),cos(bot.aiming)))
		S_StartSound(bot.mo, ring.info.seesound or sfx_thok)
		bot.weapondelay = P_RandomRange(4*TICRATE, 10*TICRATE)
		return true
	end*/
	-- Post-update variables
	ai.errorstate = false
		
	dprint('fallthrough: '..tostring(not returnVal))
	return returnVal -- Returns the value received from dirkai's thinkstate function
end)